const char* const file_ext_list[] = {
  "???",
  "ayl",
  "pt1",
  "pt2",
  "pt3",
  "stc",
  "stp",
  "asc",
  "psc",
  "sqt",
  "ay"
};

enum {
  TYPE_UNK = 0,
  TYPE_AYL,
  TYPE_PT1,
  TYPE_PT2,
  TYPE_PT3,
  TYPE_STC,
  TYPE_STP,
  TYPE_ASC,
  TYPE_PSC,
  TYPE_SQT,
  TYPE_AY,
  TYPES_ALL
};

#define BROWSER_LINES       7
#define BROWSER_WAIT_SHOW   1000/10 //delay before showing wait screen

enum {
  BROWSE_DIR,
  BROWSE_AYL
};

#define STOP_HOLD_TIME    32
#define MOVE_HOLD_TIME    16
#define MOVE_MAX_SPEED    8

#define SORT_HASH_LEN       4
#define SORT_FILES_MAX      1280  //there are 1070 entries in the Tr_Songs/Authors currently

struct sortStruct {
  char hash[SORT_HASH_LEN];
  uint16_t file_id;
};

sortStruct sort_list[SORT_FILES_MAX];

int16_t sort_list_len;

uint8_t browser_rebuild = 1;



int browser_check_ext(const char* name)
{
  for (int i = strlen(name) - 1; i >= 0; --i)
  {
    if (name[i] == '.')
    {
      for (int j = 1; j < TYPES_ALL; ++j)
      {
        if (strcasecmp(&name[i + 1], file_ext_list[j]) == 0) return j;
      }

      break;
    }
  }

  return TYPE_UNK;
}



void browser_reset_directory()
{
  strncpy(Config.active_dir, "/", sizeof(Config.active_dir) - 1);
  strncpy(Config.prev_dir, "/", sizeof(Config.prev_dir) - 1);
  strncpy(Config.ayl_file, "", sizeof(Config.ayl_file) - 1);
}



void browser_enter_directory()
{
  sd_dir.open(Config.active_dir, O_RDONLY);
  sd_file.open(&sd_dir, sort_list[Config.dir_cur].file_id, O_RDONLY);
  sd_file.getName(lfn, sizeof(lfn));
  sd_file.close();
  sd_dir.close();

  strncat(Config.active_dir, lfn, sizeof(Config.active_dir) - 1);
  strncat(Config.active_dir, "/", sizeof(Config.active_dir) - 1);
  Config.active_dir[sizeof(Config.active_dir) - 1] = 0;

  Config.dir_cur = 0;

  config_save();
}



void browser_leave_directory()
{
  memset(Config.prev_dir, 0, sizeof(Config.prev_dir));

  for (int i = strlen(Config.active_dir) - 2; i >= 0; --i)
  {
    if (Config.active_dir[i] == '/')
    {
      strncpy(Config.prev_dir, &Config.active_dir[i + 1], sizeof(Config.prev_dir) - 1);
      Config.prev_dir[strlen(Config.prev_dir) - 1] = 0; //remove trailing slash

      Config.active_dir[i + 1] = 0;

      break;
    }
  }

  Config.dir_cur = 0;

  config_save();
}



bool browser_full_path(int cur, char* path, int path_size)
{
  char temp[MAX_PATH] = {0};

  if (cur < 0 || cur >= sort_list_len) return false;

  if (sd_dir.open(Config.active_dir, O_RDONLY))
  {
    if (sd_file.open(&sd_dir, sort_list[cur].file_id, O_RDONLY))
    {
      sd_file.getName(temp, sizeof(temp));
      sd_file.close();
    }

    temp[sizeof(temp) - 1] = 0;

    sd_dir.close();

    strncpy(path, Config.active_dir, path_size - 1);
    strncat(path, temp, path_size - 1);

    return true;
  }

  path[0] = 0;
  return false;
}



int browser_build_list()
{
  memset(sort_list, 0, sizeof(sort_list));

  if (!sd_fat.begin(SD_CS, SD_SPI_SPEED))
  {
    message_box_P(PSTR(MSG_ERR_NO_CARD), MB_OK);
    return FILE_ERR_NO_CARD;
  }

  if (!sd_dir.open(Config.active_dir, O_RDONLY))
  {
    message_box_P(PSTR(MSG_ERR_CANT_OPEN_DIR), MB_OK);
    return FILE_ERR_OTHER;
  }

  int file_id = -1;

  sort_list_len = 0;

  int ptime = millis();

  while (sd_file.openNext(&sd_dir, O_RDONLY))
  {
    if (!sd_file.isHidden())
    {
      memset(lfn, 0, SORT_HASH_LEN);

      sd_file.getName(lfn, sizeof(lfn)); //full long name is needed to be able to check the extension

      if (sd_file.isSubDir() || (sd_file.isFile() && (browser_check_ext(lfn) != TYPE_UNK)))
      {
        if (sort_list_len >= SORT_FILES_MAX)
        {
          sd_file.close();
          sd_dir.close();
          sort_list_len = 0;
          char str[128];
          snprintf(str, sizeof(str), "Too many files\nin a folder\n(%u max)", SORT_FILES_MAX);
          message_box(str, MB_OK);
          return FILE_ERR_OTHER;
        }

        if (sd_file.isSubDir())
        {
          if (strcasecmp(Config.prev_dir, lfn) == 0) file_id = sd_file.dirIndex();
        }

        sort_list[sort_list_len].file_id = sd_file.dirIndex();

        for (int i = 0; i < SORT_HASH_LEN; ++i)
        {
          if (lfn[i] >= 'a' && lfn[i] <= 'z') lfn[i] -= 32;  //to upper case just in case
          if (lfn[i] == '_') lfn[i] = 0x20; //put underscore names to the top
        }

        if (sd_file.isSubDir())
        {
          sort_list[sort_list_len].hash[0] = 1;
          memcpy(&sort_list[sort_list_len].hash[1], lfn, SORT_HASH_LEN - 1);
        }
        else
        {
          memcpy(sort_list[sort_list_len].hash, lfn, SORT_HASH_LEN);
        }

        ++sort_list_len;
      }
    }

    sd_file.close();

    if ((millis() - ptime) >= BROWSER_WAIT_SHOW)
    {
      ptime = millis();
      wait_screen();
    }
  }

  sd_dir.close();

  if (sort_list_len == 0)
  {
    message_box_P(PSTR(MSG_ERR_NO_FILES), MB_OK);
    return FILE_ERR_OTHER;
  }

  //now sort the list for a convenient alphanumeric ordered display

  //insertion sort

  int i = 1;

  while (i < sort_list_len)
  {
    int j = i;

    while (j > 0)
    {
      int carry = 1;

      for (int k = SORT_HASH_LEN - 1; k >= 0; --k)
      {
        carry = ((sort_list[j].hash[k] - sort_list[j - 1].hash[k] - carry) < 0) ? 1 : 0;
      }

      if (!carry) break;

      sortStruct temp;

      memcpy(&temp, &sort_list[j], sizeof(sortStruct));
      memcpy(&sort_list[j], &sort_list[j - 1], sizeof(sortStruct));
      memcpy(&sort_list[j - 1], &temp, sizeof(sortStruct));

      --j;
    }

    ++i;

    if ((millis() - ptime) >= BROWSER_WAIT_SHOW)
    {
      ptime = millis();
      wait_screen();
    }
  }

  //or bubblesort
  /*
    for (int i = 0; i < sort_list_len - 1; ++i)
    {
      for (int j = 0; j < sort_list_len - i - 1; ++j)
      {
        //if arr j > arr j+1 swap

        int carry = 1;

        for (int k = SORT_HASH_LEN - 1; k >= 0; --k)
        {
          carry = ((sort_list[j].hash[k] - sort_list[j + 1].hash[k] - carry) < 0) ? 1 : 0;
        }

        if (!carry)
        {
          sortStruct temp;

          memcpy(&temp, &sort_list[j], sizeof(sortStruct));
          memcpy(&sort_list[j], &sort_list[j + 1], sizeof(sortStruct));
          memcpy(&sort_list[j + 1], &temp, sizeof(sortStruct));
        }
      }
    }
  */

  for (int i = 0; i < sort_list_len; ++i)
  {
    if (sort_list[i].file_id == file_id)
    {
      Config.dir_cur = i;
      break;
    }
  }

  if (Config.dir_cur < 0 || Config.dir_cur >= sort_list_len) Config.dir_cur = 0;

  return FILE_ERR_NONE;
}



bool browser_move_cur(int dir, bool loop)
{
  bool done = false;

  frame_cnt = 0;
  Config.dir_cur += dir;

  if (loop)
  {
    if (Config.dir_cur < 0) Config.dir_cur = sort_list_len - 1;
    if (Config.dir_cur >= sort_list_len) Config.dir_cur = 0;
  }
  else
  {
    if (Config.dir_cur < 0)
    {
      Config.dir_cur = 0;
      done = true;
    }
    if (Config.dir_cur >= sort_list_len)
    {
      Config.dir_cur = sort_list_len - 1;
      done = true;
    }
  }

  return done;
}



void browser_shuffle_cur()
{
  frame_cnt = 0;

  int prev_cur = Config.dir_cur;

  while (1)
  {
    Config.dir_cur = rand() % sort_list_len;

    if (Config.dir_cur != prev_cur) break;

    yield();
  }
}



void browser_dir_draw_begin(int id)
{
  sd_dir.open(Config.active_dir, O_RDONLY);
}

void browser_dir_draw_item(int sx, int sy, int id)
{
  sd_file.open(&sd_dir, sort_list[id].file_id, O_RDONLY);
  sd_file.getName(lfn, sizeof(lfn));
  sd_file.close();

  lfn[sizeof(lfn) - 1] = 0;

  char prefix = 0;

  if (sort_list[id].hash[0] == 1) prefix = '\2';
  if (browser_check_ext(lfn) == TYPE_AYL) prefix = '\1';

  if (prefix)
  {
    display_char(sx, sy, prefix);
    display_str(sx + 8, sy, lfn);
  }
  else
  {
    display_str(sx, sy, lfn);
  }
}

void browser_dir_draw_end()
{
  sd_dir.close();
}


void browser_ayl_draw_begin(int id)
{
  playlist_open(Config.ayl_file, id);
}

void browser_ayl_draw_item(int sx, int sy, int id)
{
  playlist_iterate(lfn, sizeof(lfn));
  playlist_file_name(lfn, sizeof(lfn));

  display_str(sx, sy, lfn);
}

void browser_ayl_draw_end()
{
  playlist_close();
}



//returns cursor position 0..sort_list_len-1 in sort_list or playlist to play
//-1 on a long STOP hold for config screen
//-2 on short STOP press to go back

int browser_screen(int mode)
{
  PGM_P header_dir = PSTR("Files");
  PGM_P header_ayl = PSTR("Playlist");

  while (1)
  {
    if (mode == BROWSE_DIR)
    {
      if (browser_rebuild)
      {
        browser_rebuild = 0;

        int ret = browser_build_list();

        if (ret != FILE_ERR_NONE)
        {
          if (ret != FILE_ERR_NO_CARD)
          {
            browser_leave_directory();
          }
          else
          {
            browser_reset_directory();
          }

          browser_rebuild = 1;
          continue;
        }
      }
    }
    else
    {
      if (!playlist_open(Config.ayl_file, 0)) return -1;

      sort_list_len = 0;

      while (playlist_iterate(lfn, sizeof(lfn)))
      {
        ++sort_list_len;
      }

      playlist_close();
    }

    bool update = true;
    bool update_list = true;
    int speedup = 0;
    int cur_speed = 1;

    int stop_down = 0;
    frame_cnt = 0;

    display_clear();

    int sd_check_timer = 0;

    while (1)
    {
      //checking if there is a card inserted in case it gets pulled out

      if (display_is_active())
      {
        if (!sd_check_timer)
        {
          sd_check_timer = 0;

          if (!sd_fat.begin(SD_CS, SD_SPI_SPEED))
          {
            browser_rebuild = true;
            break;
          }
        }

        ++sd_check_timer;

        if (sd_check_timer >= 1 * 60) sd_check_timer = 0;
      }

      if (!(frame_cnt & 7)) update = true;

      if (pad_force_redraw()) update_list = true;

      if (display_is_active() && (update || update_list))
      {
        int id = Config.dir_cur - BROWSER_LINES / 2;

        if (id >= sort_list_len - BROWSER_LINES) id = sort_list_len - BROWSER_LINES;
        if (id < 0) id = 0;

        display_mode(DM_AND);
        display_rect(0, 0, VIEW_WDT, 8);
        display_mode(DM_OR);

        display_charge(111, 0);

        PGM_P header_str = (mode == BROWSE_DIR) ? header_dir : header_ayl;

        display_rect(0, 0, display_str_len_P(header_str) + 8, 8);
        display_mode(DM_AND);
        display_str_P(4, 1, header_str);
        display_mode(DM_OR);

        int sx = 6;
        int sy = 9;

        if (update_list)
        {
          if (mode == BROWSE_DIR)
          {
            browser_dir_draw_begin(id);
          }
          else
          {
            browser_ayl_draw_begin(id);
          }

          display_mode(DM_AND);
          display_rect(sx, sy, VIEW_WDT - sx, VIEW_HGT - sy);
          display_mode(DM_OR);
        }

        for (int i = 0; i < BROWSER_LINES; ++i)
        {
          if (id >= 0 && id < sort_list_len)
          {
            if ((!(frame_cnt & 8) && (Config.dir_cur == id)))
            {
              display_mode(DM_OR);
              display_char(0, sy, 0x00);
            }
            else
            {
              display_mode(DM_AND);
              display_char(0, sy, 0x00);
              display_mode(DM_OR);
            }

            if (update_list)
            {
              if (mode == BROWSE_DIR)
              {
                browser_dir_draw_item(sx, sy, id);
              }
              else
              {
                browser_ayl_draw_item(sx, sy, id);
              }
            }

            sy += 8;
          }

          ++id;
        }

        if (update_list)
        {
          if (mode == BROWSE_DIR)
          {
            browser_dir_draw_end();
          }
          else
          {
            browser_ayl_draw_end();
          }
        }

        update = false;
        update_list = false;
      }

      display_update();

      wait_frame();

      pad_check();

      if (pad_state & (PAD_FWD | PAD_REW))
      {
        ++speedup;

        if (speedup > MOVE_HOLD_TIME)
        {
          speedup = 1;
          ++cur_speed;
          if (cur_speed > MOVE_MAX_SPEED) cur_speed = MOVE_MAX_SPEED;
        }
      }
      else
      {
        speedup = 0;
        cur_speed = 1;
      }

      if (pad_state_rep & PAD_REW)
      {
        sound_play(SFX_MOVE);
        update_list = true;
        browser_move_cur(-cur_speed, true);
      }

      if (pad_state_rep & PAD_FWD)
      {
        sound_play(SFX_MOVE);
        update_list = true;
        browser_move_cur(cur_speed, true);
      }

      if (pad_state_down & PAD_PLAY)
      {
        if (mode == BROWSE_DIR)
        {
          if (sort_list[Config.dir_cur].hash[0] == 1)  //directory
          {
            sound_play(SFX_SELECT);
            browser_enter_directory();
            browser_rebuild = 1;
            break;
          }
        }

        return Config.dir_cur;
      }

      if (pad_state & PAD_STOP)
      {
        ++stop_down;

        if (stop_down >= STOP_HOLD_TIME)
        {
          sound_play(SFX_SELECT);
          return -1;
        }
      }
      else
      {
        stop_down = 0;
      }

      if (pad_state_up & PAD_STOP)
      {
        sound_play(SFX_CANCEL);

        if (mode == BROWSE_DIR)
        {
          browser_leave_directory();
          browser_rebuild = 1;
        }

        break;
      }
    }

    if (mode == BROWSE_AYL) break;
  }

  return -2;
}
