#define SPEC_BANDS        42
#define SPEC_HEIGHT       24
#define SPEC_RANGE        1000
#define SPEC_SX           0
#define SPEC_SY           28
#define SPEC_BAND_WIDTH   3
#define SPEC_DECAY        3

#define SPEC_CHA_COL      0
#define SPEC_CHB_COL      0
#define SPEC_CHC_COL      0
#define SPEC_NSE_COL      3
#define SPEC_ENV_COL      2


int16_t spec_levels[SPEC_BANDS];
uint8_t spec_colors[SPEC_BANDS];




void spec_init()
{
  for (int i = 0; i < SPEC_BANDS; ++i)
  {
    spec_levels[i] = 0;
    spec_colors[i] = 0;
  }
}



void spec_add(int hz, int level, uint8_t color)
{
  int i, off;
  const int16_t curve[5] = {SPEC_HEIGHT / 10, SPEC_HEIGHT / 5, SPEC_HEIGHT / 2, SPEC_HEIGHT / 5, SPEC_HEIGHT / 10};

  if (hz)
  {
    off = hz / (SPEC_RANGE / SPEC_BANDS) - 2;

    if (off > SPEC_BANDS - 1) off = SPEC_BANDS - 1;

    for (i = 0; i < 5; ++i)
    {
      if (off >= 0 && off < SPEC_BANDS)
      {
        spec_levels[off] += curve[i] * level / 16;

        if (spec_levels[off] > SPEC_HEIGHT) spec_levels[off] = SPEC_HEIGHT;

        spec_colors[off] = color;
      }

      ++off;
    }
  }
}



void spec_update(bool is_ts)
{
  for (int i = 0; i < SPEC_BANDS; ++i)
  {
    spec_levels[i] -= SPEC_DECAY;
    if (spec_levels[i] < 0) spec_levels[i] = 0;
  }

  uint8_t* reg;

  for (int chip = 0; chip < (is_ts ? 2 : 1); ++chip)
  {
    if (!chip) reg = hway_reg_1; else reg = hway_reg_2;

    int period;

    if (!(reg[7] & 0x08))
    {
      period = (SPEC_RANGE / 2 + SPEC_RANGE / 3) * (31 - reg[6]) / 32;
      spec_add(period, reg[8] & 15, SPEC_NSE_COL);
    }
    if (!(reg[7] & 0x10))
    {
      period = (SPEC_RANGE / 2 + SPEC_RANGE / 3) * (31 - reg[6]) / 32;
      spec_add(period, reg[9] & 15, SPEC_NSE_COL);
    }
    if (!(reg[7] & 0x20))
    {
      period = (SPEC_RANGE / 2 + SPEC_RANGE / 3) * (31 - reg[6]) / 32;
      spec_add(period, reg[10] & 15, SPEC_NSE_COL);
    }
    if (!(reg[7] & 0x01))
    {
      period = reg[0] + reg[1] * 256;
      if (period) spec_add(Config.ay_clock / 16 / period, reg[8] & 15, SPEC_CHA_COL);
    }
    if (!(reg[7] & 0x02))
    {
      period = reg[2] + reg[3] * 256;
      if (period) spec_add(Config.ay_clock / 16 / period, reg[9] & 15, SPEC_CHB_COL);
    }
    if (!(reg[7] & 0x04))
    {
      period = reg[4] + reg[5] * 256;
      if (period) spec_add(Config.ay_clock / 16 / period, reg[10] & 15, SPEC_CHC_COL);
    }
    if ((reg[8] & 0x10) || (reg[9] & 0x10) || (reg[10] & 0x10))
    {
      period = reg[11] + reg[12] * 256;
      if (period) spec_add(Config.ay_clock / 16 / 16 / period, 12, SPEC_ENV_COL);
    }
  }
}
